\documentclass{article}
\include{/hugs/lib/activex/lhs}

\begin{document}

\begin{notice}
Copyright (c) 1998,  Daan Leijen, leijen@@fwi.uva.nl

This module is part of HaskellDirect (H/Direct).
See @notice.txt@ for details and conditions of use.
\end{notice}

> module Com (

Basic interfaces.

>       (#), (##), with, withObject_, withObject, withMethod, withMethod_,
>       interfaceNULL, iidNULL,
>       Interface,
>       Unknown, IUnknown, iidIUnknown,
>       queryInterface, addRef, release,

COM Errors.

>       HRESULT, checkHR, succeededHR, failedHR, coFailHR, stringFromHR,
>       coFail, isCoError, coGetErrorString, 
>       coAssert, coOnFail,

COM library functions.

>       coRun, coPerformIO,
>       coCreateObject,
>       coGetObject,
>       coGetActiveObject,
>       coGetFileObject,
>       coCreateInstance,

GUID management.

>       ProgID,
>       IID, CLSID, GUID,
>       mkIID, mkCLSID, mkGUID,
>       createIID, createCLSID, createGUID,
>       stringFromIID, stringFromCLSID, stringFromGUID,
>       clsidFromProgID, progIDFromCLSID,

Process contexts.

>       CLSCTX(..),
>       -- LocalProcess, inProcess, ServerProcess, AnyProcess,

Misc.

>       true, false, boolFromInt, intFromBool,
>       printMessage, putMessage, playSound,

>       enumClone, enumNext, enumSkip, enumReset,

>       comRun, comCreateObject

>       ) where
>

> import PointerPrim (primPointerCheck)
> import Pointer ( Pointer, pointerNULL, allocMemory )
> import Marshall ( stackString, stackStringLen, stackWideString,
>                   stringFromWide, readStringAt
>                )
> import ComPrim
> import IO      (ioeGetErrorString,isUserError)
> import IOExts  (unsafePerformIO)
> import LibMonad (seqs,(@@))
> import LibIO   (always,onfail)


Method call operators. For example:
@genie \# move 10 10@. Use @\#\#@ to call methods on
IO returned interfaces:
@agent \# characters \#\# item "genie" \#\# showUp 0@.

> infixl 1 #
> infixl 2 ##

> obj # method          = method obj
> m1 ## m2              = \x -> (m1 x >>= m2)


> comRun                = coRun
> comCreateObject       = coCreateObject


> instance MonadZero IO where
>   zero    = fail (userError "IO pattern match failed")

@withObject@ applies every method in a list to an 
object: @withObject genie [showUp, speak "hi", hide]@.
@withMethod@ applies every argument in a list to a
method: @genie # withMethod speak ["hello", "world"]@.

> with  = withObject_
>
> withObject_ :: IUnknown a -> [IUnknown a -> IO b] -> IO ()
> withObject_ obj       = sequence .  map (obj#) 

> withMethod_ :: (a -> IUnknown b -> IO c) -> [a] -> IUnknown b -> IO ()
> withMethod_ method args obj
>                       = sequence $ map (\x -> obj # method x) args

> withObject :: IUnknown a -> [IUnknown a -> IO b] -> IO [b]
> withObject obj        = accumulate . map (obj#)
>
> withMethod :: (a -> IUnknown b -> IO c) -> [a] -> IUnknown b -> IO [c]
> withMethod method args obj 
>                       = accumulate $ map (\x -> obj # method x) args

HRESULT is the COM way of returning errors. The following
functions integrate it with IO.

> type HRESULT          = Int

> checkHR :: HRESULT -> IO ()
> checkHR hr
>       | failedHR hr   = coFailHR hr
>       | otherwise     = return ()
>
> succeededHR :: HRESULT -> Bool
> succeededHR           = not . failedHR
>
> coFailHR :: HRESULT -> IO a
> coFailHR             = coFail @@ stringFromHR
>
> stringFromHR :: HRESULT -> IO String
> stringFromHR hr
>       = do pstr <- primStringFromHR hr  -- static memory
>            readStringAt pstr


Many COM interfaces use integers for booleans.

> intFromBool b         = if b then true else false
> boolFromInt i         = if (i == false) then False else True
>
> true,false :: Int
> true                  = (-1)
> false                 = 0

The normal 'user error' is used to pass COM errors for now.

> coAssert :: Bool -> String -> IO ()
> coAssert True msg    = return ()
> coAssert False msg   = coFail msg

> coOnFail :: IO a -> String -> IO a
> coOnFail io msg      = io `catch` \err -> 
>                         coFail (msg ++ ": " ++ (coGetErrorString err))

> coPrefix              = "Com error: "
> userPrefix            = "User error: "
> prefixLen             = length userPrefix + length coPrefix
>
> coFail msg            = fail (userError (coPrefix ++ msg))
>
> isCoError err         = (isUserError err)  &&
>                         (prefix == userPrefix ++ coPrefix)
>                       where
>                         prefix    = take prefixLen (ioeGetErrorString err)
>
> coGetErrorString e    | isCoError e    = drop prefixLen (ioeGetErrorString e)
>                       | otherwise      = drop (length userPrefix) (ioeGetErrorString e)


COM initialize/uninitialize. 
Uninitialize is commented out since I got strange errors
(seems you can't reinitalize once uninitialized) but 
I think it should be in.

> coRun :: IO a -> IO ()
> coRun io
>       = do coInitialize
>            coPerformIO io  --`always` coUnInitialize
>            primPointerCheck
>            return ()

> coPerformIO :: IO a -> IO a
> coPerformIO io
>       = io `catch` \err ->
>          do putMessage (coGetErrorString err)
>             fail err

> coUnsafePerformIO :: IO a -> a
> coUnsafePerformIO
>       = unsafePerformIO . coPerformIO

Show a nice message box.

> printMessage :: Show a => a -> IO ()
> printMessage x
>       = putMessage (show x)
>
> putMessage msg
>       = stackString msg $ \pmsg ->
>         messageBox pmsg


Some standard COM interfaces.

> iidNULL :: IID ()
> iidNULL               = mkIID "{00000000-0000-0000-0000-000000000000}"

> iidIUnknown :: IID (IUnknown Interface)
> iidIUnknown           = mkIID "{00000000-0000-0000-C000-000000000046}"

@IPersistFile@ is used in the @getObject@ function.

> iidIPersistFile :: IID (IPersistFile Interface)
> iidIPersistFile       = mkIID "{0000010B-0000-0000-C000-000000000046}"


The interface data type.

> interfaceNULL :: IUnknown a
> interfaceNULL         = Unknown pointerNULL
>
> instance Eq (Unknown a) where
>   iface1 == iface2    = comEqual iface1 iface2
>
> instance Show (Unknown a) where
>   showsPrec d iface   = shows "<com object>"

Objects can be identified with a programmatic identifier too: @"Agent.Server"@.

> type ProgID   = String

Create a fresh object.
@coCreateObject@ creates an object from its progID: 
   @coCreateObject "Agent.Server" iidIAgent@.

> coCreateObject :: ProgID -> IID (IUnknown a) -> IO (IUnknown a)
> coCreateObject progid iid
>       = do clsid  <- clsidFromProgID progid
>            coCreateInstance clsid interfaceNULL AnyProcess iid

Get Object from File \& ProgID.
@coGetObject@ creates an object from its progID and initializes it with a given file:
@coGetObject "spreadsheet.exl" "Excel.Application" iidIApplication@. 
If the filename is empty, @coGetObject@ calls @coGetActiveObject@.


> coGetObject :: String -> ProgID -> IID (IUnknown a) -> IO (IUnknown a)
> coGetObject "" progid iid
>       = coGetActiveObject progid iid
>
> coGetObject fname progid iid
>       = do pf <- coCreateObject progid iidIPersistFile
>            stackWideString fname $ \pfname ->
>             do persistfileLoad pf pfname
>                pf # queryInterface iid

Get Active Object.
@coGetActiveObject@ tries to connect to an already running instance of the coponent:
@coGetActiveObject "Word.Document" iidIWordDocument@.

> coGetActiveObject :: ProgID -> IID (IUnknown a) -> IO (IUnknown a)
> coGetActiveObject progid iid
>       = do clsid   <- clsidFromProgID progid
>            iface   <- primGetActiveObject clsid
>                       `coOnFail` ("Could not connect to component '" ++ progid ++ "'")
>            iface # queryInterface iid

Get File Object.
@coGetFileObject@ opens a file or url and loads the associated or persistent object in it:
@coGetFileObject "spreadsheet.spd" iidIWorkBook@.

> coGetFileObject :: String -> IID (IUnknown a) -> IO (IUnknown a)
> coGetFileObject fname iid
>       = stackWideString fname $ \pfname ->
>         do bindObject pfname iid

Create instance.
@coCreateInstance@ is the basic com way of creating components. It takes
a CLSID, an interface to aggregate on, a process context and an IID to create an object:
@coCreateInstance clsidAgentServer interfaceNULL LocalProcess iidIAgent@.

> coCreateInstance :: CLSID -> IUnknown b ->
>                                 CLSCTX -> IID (IUnknown a) -> IO (IUnknown a)
> coCreateInstance clsid inner context iid
>       = primCreateInstance clsid inner (fromEnum context) iid

IID, CLSID, GUID's.

> mkIID :: String -> IID a
> mkIID                 = mkGUID
> mkCLSID               = mkGUID
>
> mkGUID :: String -> GUID
> mkGUID str
>       = coUnsafePerformIO $
>         stackWideString str $ \pstr ->
>         do primGUIDFromString pstr
>
> createIID :: IO (IID a)
> createIID             = createGUID
> createCLSID           = createGUID
>
> stringFromIID :: IID a -> IO String
> stringFromIID         = stringFromGUID
> stringFromCLSID       = stringFromGUID
>
> stringFromGUID :: GUID -> IO String
> stringFromGUID guid
>       = do pstr <- primStringFromGUID guid
>            readStringAt pstr
>
> instance Show GUID where
>   showsPrec d guid
>       = showString (coUnsafePerformIO (stringFromGUID guid))

> clsidFromProgID :: ProgID -> IO CLSID
> clsidFromProgID progid
>       = stackString progid $ \pprogid ->
>         do primCLSIDFromProgID pprogid
>               `coOnFail` ("Component '" ++ progid ++ "' is unknown")
>
> progIDFromCLSID :: CLSID -> IO ProgID
> progIDFromCLSID clsid
>       = stackStringLen 255 "" $ \pstr ->
>         do pwide <- primProgIDFromCLSID clsid
>            stringFromWide pwide 255 pstr
>            readStringAt pstr


Process contexts.

> data CLSCTX = CLSCTX_INPROC_SERVER |CLSCTX_INPROC_HANDLER
>                 |CLSCTX_LOCAL_SERVER |CLSCTX_INPROC_SERVER16
>                 |CLSCTX_REMOTE_SERVER |CLSCTX_INPROC_HANDLER16
>                 |CLSCTX_INPROC_SERVERX86 |CLSCTX_INPROC_HANDLERX86
>                 |LocalProcess
>                 |InProcess
>                 |ServerProcess
>                 |AnyProcess
>               deriving (Eq ,Ord ,Show)
>
> instance Enum CLSCTX where
>   toEnum x = case x of
>                1 -> CLSCTX_INPROC_SERVER
>                2 -> CLSCTX_INPROC_HANDLER
>                8 -> CLSCTX_INPROC_SERVER16
>                16 -> CLSCTX_REMOTE_SERVER
>                32 -> CLSCTX_INPROC_HANDLER16
>                64 -> CLSCTX_INPROC_SERVERX86
>                128 -> CLSCTX_INPROC_HANDLERX86
>                0x04 -> LocalProcess
>                0x0B -> InProcess
>                0x0D -> ServerProcess
>                4    -> CLSCTX_LOCAL_SERVER
>                _    -> AnyProcess
>
>   fromEnum x = case x of
>                  CLSCTX_INPROC_SERVER -> 1
>                  CLSCTX_INPROC_HANDLER -> 2
>                  CLSCTX_LOCAL_SERVER -> 4
>                  CLSCTX_INPROC_SERVER16 -> 8
>                  CLSCTX_REMOTE_SERVER -> 16
>                  CLSCTX_INPROC_HANDLER16 -> 32
>                  CLSCTX_INPROC_SERVERX86 -> 64
>                  CLSCTX_INPROC_HANDLERX86 -> 128
>                  LocalProcess   -> 0x04
>                  InProcess      -> 0x0B
>                  ServerProcess  -> 0x0D
>                  AnyProcess     -> 0x0F

Generic IEnum* entry points

> enumReset :: IUnknown a -> IO ()
> enumReset ip = primEnumReset ip
> 
> enumSkip  :: Int -> IUnknown a -> IO ()
> enumSkip i ip = primEnumSkip ip i
> 
> enumNext :: Int -> Pointer b -> IUnknown a -> IO Int
> enumNext celt ptr ip = primEnumNext ip celt ptr
> 
> enumClone :: IUnknown a -> IO (IUnknown a)
> enumClone ip = primEnumClone ip

Misc.

> playSound :: String -> IO ()
> playSound fname
>       = stackString fname $ \pstr ->
>         primPlaySound pstr



\end{document}

