\documentclass{article}
\include{lhs}

\begin{document}

\begin{notice}
Copyright (c) 1998,  Daan Leijen, leijen@@fwi.uva.nl

This module is part of HaskellScript (HScript).
See @notice.txt@ for details and conditions of use.
\end{notice}

> module HSPrim (
>                 externItem,
>                 rootItem,
>                 onEventID,
>                 onEvent
>               ) where

> import StdDIS           
> import PointerPrim      (unsafePointerFromInt,unsafeIntFromPointer)
> import ComPrim          (unsafeMkInterface,unsafeGetInterfacePointer)
> import Pointer          (Pointer,pointerNULL)
> import IOExts           (unsafePerformIO)
> import Com
> import Automation
> import LibMonad       ((@@))
> import LibIO          (mapio)
> import LibRef


Create an interface pointer from an address.

> unsafeMkDispatch :: Int -> IDispatch a
> unsafeMkDispatch
>       = unsafeMkInterface . unsafePointerFromInt


\textit{ScriptEngine Entry:} Initialize

> initialize :: IO ()
> initialize      = do resetItems       -- hmmmm ?
>                      resetEvents
>                      return ()

Top Level Objects, \textit{ScriptEngine entry:} addItem

> refItems :: Ref [(String,IDispatch a)]
> refItems
>       = unsafePerformIO (newRef [])
>
> resetItems :: IO ()
> resetItems
>       = setRef refItems []
>
> addItem :: String -> Int -> IO ()
> addItem name addr
>       = do let obj = unsafeMkDispatch addr
>            updateRef refItems (replaceItem obj)
>       where
>         replaceItem obj []         = [(name, obj)]
>         replaceItem obj (x:xs)
>               | fst x == name  = (name, obj):xs
>               | otherwise      = x : replaceItem obj xs
>
> externItem :: String -> IO (IDispatch a)
> externItem name
>       = do is <- getRef refItems
>            case (lookup name is) of
>              Just disp   -> return disp
>              Nothing     -> comFail ((show name) ++ ": unknown item")
>
> rootItem :: IO (IDispatch a)
> rootItem
>       = do is <- getRef refItems
>            case is of
>              []               -> comFail ("rootItem: no root object available")
>              ((name,root):xs) -> return root

The interfaces exposed from the script engine to haskell.

> data HaskellEngine  = HaskellEngine
> type IHaskellEngine = IDispatch HaskellEngine

> data HaskellSink    = HaskellSink
> type IHaskellSink   = IDispatch HaskellSink


These function connect haskell to the engine:

> engineConnect :: IDispatch a -> IHaskellEngine -> IO IHaskellSink
> engineConnect obj       = function1 "Connect" [inIDispatch obj] outIDispatch
>
> sinkEventID :: String -> IHaskellSink -> IO Int
> sinkEventID event       = function1 "EventID" [inString event] outInt



The engine connection, \textit{ScriptEngine entry:} setScriptEngine

> scriptEngine :: Ref (Pointer Interface)
> scriptEngine
>       = unsafePerformIO (newRef pointerNULL)
>
> setScriptEngine :: Int -> IO ()
> setScriptEngine addr
>       = setRef scriptEngine (unsafePointerFromInt addr)
>
> getScriptEngine :: IO (Maybe IHaskellEngine)
> getScriptEngine
>       = do p <- getRef scriptEngine
>            if (p == pointerNULL)
>             then return Nothing
>             else return (Just (unsafeMkInterface p))

\section*{Events}
\textit{ScriptEngine Entry:} @resetEvent@, @execEvent@.

The event map is a list of triples.
The triples consist of an object (as a @pointer@),
a sink that connects this object and
a list of tuples: the event (disp)id and the associated action

> type EventMap   = [(Int, IDispatch Interface, IHaskellSink, [(DISPID, IO ())])]
>
> eventMap        :: Ref EventMap
> eventMap        = unsafePerformIO (newRef [])
>
> resetEvents     = setRef eventMap []

> baseIDispatch :: IDispatch a -> IDispatch Interface
> baseIDispatch   = unsafeMkInterface . unsafeGetInterfacePointer

Connect a certain object to an event sink

> connect :: IDispatch a -> IO IHaskellSink
> connect disp    = do mbEngine <- getScriptEngine
>                      case mbEngine of
>                        Nothing      -> return interfaceNULL
>                        Just engine  -> engine # engineConnect disp

@lookupSink@ tries to find a certain object in the eventMap.

> lookupSink p []
>       = Nothing
> lookupSink p ((objid,obj,sink,events):xs)
>       | p == objid    = Just (sink,events)
>       | otherwise     = lookupSink p xs

@removeSink@ tries to find a certain object in the eventMap
and removes it.

> removeSink p []
>       = (Nothing, [])
> removeSink p (x@(objid,obj,sink,events):xs)
>       | p == obj      = (Just (sink,events), xs)
>       | otherwise     = let (y,ys) = removeSink p xs
>                         in  (y,x:xs)

@onEventID@ and @OnEvent@ connect a certain object to an event
by dispid or name.

> onEventID :: DISPID -> IO () -> IDispatch a -> IO ()
> onEventID dispid action obj'
>       = do let obj        =  baseIDispatch obj'
>                objid      =  unsafeIntFromPointer (unsafeGetInterfacePointer obj)
>            events         <- getRef eventMap
>            ((sink,es),xs) <- case (removeSink obj events) of
>                                (Nothing,_)  -> do sink <- obj # connect
>                                                   return ((sink,[]),events)
>                                (Just x,xs)  -> return (x,xs)
>            setRef eventMap ((objid,obj,sink,(dispid,action):es) :xs)

> onEvent :: String -> IO () -> IDispatch a -> IO ()
> onEvent event action obj'
>       = do let obj        =  baseIDispatch obj'
>                objid      = unsafeIntFromPointer (unsafeGetInterfacePointer obj)
>            events         <- getRef eventMap
>            ((sink,es),xs) <- case (removeSink obj events) of
>                                (Nothing,_) -> do sink <- obj # connect
>                                                  return ((sink,[]),events)
>                                (Just x,xs) -> return (x,xs)
>            dispid         <- sink # sinkEventID event
>            setRef eventMap ((objid,obj,sink,(dispid,action):es) :xs)


 onEvent0 :: String -> IO () -> IDispatch a -> IO ()
 onEvent1 :: OutVariant a => String -> (a -> IO ()) -> IDispatch a -> IO ()
 onEvent2 :: (OutVariant a,OutVariant b) => String -> (a -> b -> IO ()) -> IDispatch a -> IO ()


This is the entry point of an event.

> execEvent :: Int -> Int -> IO ()
> execEvent addr dispid
>       = do --pobj   <- unsafePointerFromInt addr
>            events <- getRef eventMap
>            case (lookupSink addr events) of
>              Nothing         -> return ()
>              Just (sink,es)  -> case (lookup dispid es) of
>                                   Nothing       -> return ()
>                                   Just action   -> action

\end{document}
