\documentclass{article}
\include{/hugs/lib/activex/lhs}

\begin{document}

\begin{notice}
Copyright (c) 1998,  Daan Leijen, leijen@@fwi.uva.nl

This module is part of HaskellDirect (H/Direct).
See @notice.txt@ for details and conditions of use.
\end{notice}

> module Pointer  ( Pointer,
>                   pointerNULL, pointerPlus,
>                   setFinalizer, setRootPointer, getRootPointer,
>                   allocMemory, stackFrame,
>
>                   pointerInc,pointerDec,
>                   writeSeqAt, writeSeqAtDec,
>                   readSeqAt, readSeqAtWhile,
>
>                   makePointer,
>                   makeCPointer, makeInterfacePointer, makeNoFreePointer,
>                   makeForeignPointer,
>
>                   freeMemory, freeBSTR,
>                   freeWith, freeWithC
>
>                 ) where

> import IOExts   ( unsafePerformIO )
> import Foreign  (ForeignObj,makeForeignObj,writeForeignObj)
> import StdDIS   ( Addr )
> import LibMonad ( (@@), seqs, maplRev, maplWhile )
> import LibIO    ( mapio, always )
> import PointerPrim

The @Pointer@ data type is a typed pointer to something
outside haskell. On garbage collection, it will automatically
call a supplied finalizer. Pointers are created with @makePointer@
which takes a (haskell) finalizer routine and an address.
Since foreign objects are not yet completely supported,
it is not possible to change the finalizer routine. 
The finalizer routine is also
not completely supported and is an addr to a C finalizer routine
at the moment. Special creation routines are:
@makeInterfacePointer@ calls @release@ on a COM object,
@makeCPointer@ calls @free@ (or @CoTaskMemFree@) on a C pointer,
and @makeNoFreePointer@ doesn't call a finalizer.

Pointers can be freed manually if it is statically known that
some memory can be freed. Use @freePointer@ with a finalizer routine
or call the specialized @freeMemory@. (The pointer will NOT
be freed again on GC).

@Pointer@ simply a typed wrapper around the @Ptr@ type.

@> newtype Pointer a    = Pointer Ptr

> type Finalizer a     = Addr   -- For Now, should be: Pointer a -> IO ()
>
> makePointer          :: Finalizer a -> Addr -> IO (Pointer a)
> makeInterfacePointer :: Addr -> IO (Pointer a)
> makeNoFreePointer    :: Addr -> IO (Pointer a)
> makeCPointer         :: Addr -> IO (Pointer a)
>
> makeInterfacePointer    = makePointer finalFreeObject
> makeNoFreePointer       = makePointer finalNoFree
> makeCPointer            = makePointer finalFreeMemory
> makePointer finalizer   = mapio Pointer . makePtr finalizer

> makeForeignPointer :: ForeignObj -> (Pointer a)
> makeForeignPointer f    = let p = Ptr f p
>                           in  Pointer p

The finalizer can dynamically be changed while running the
program (is this still ref. transparent ?). Not supported
yet.

> setFinalizer :: Finalizer a -> Pointer a -> IO ()
> setFinalizer free (Pointer ptr)
>       = setFinalizerPtr free ptr      -- unsupported yet

Using the @setRootPointer@ function, it is possible to
create pointers inside the memory area of another pointer;
the inner pointer will keep a reference to the root pointer,
so the outer pointer will not be garbage collected while
one of them is still live. Of course, the inner pointer
shouldn't have a finalizer.

> setRootPointer :: Pointer b -> Pointer a -> Pointer a
> setRootPointer (Pointer root) (Pointer ptr)
>       = Pointer (setRootPtr root ptr)
>
> getRootPointer :: Pointer a -> Pointer b
> getRootPointer (Pointer ptr)
>       = Pointer (getRootPtr ptr) 

@pointerPlus@ sets the root reference when doing pointer
arithmetic and makes sure the inner pointer won't be finalized.

> pointerPlus :: Pointer a -> Int -> Pointer a
> pointerPlus p offset
>       = setRootPointer (getRootPointer p) (primPointerPlus p offset)
>
> instance Eq (Pointer a) where
>       p1 == p2   = pointerEqual p1 p2

Helpers.

> pointerInc i p        = pointerPlus p i
> pointerDec d p        = pointerPlus p (-d)
>
> writeSeqAt :: Int -> [Pointer a -> IO ()] -> Pointer a -> IO ()
> writeSeqAt size ws
>       = seqs . zipWith ($) ws . iterate (pointerInc size)
>
> writeSeqAtDec :: Int -> [Pointer a -> IO ()] -> Pointer a -> IO ()
> writeSeqAtDec size ws
>       = seqs . zipWith ($) ws .
>          iterate (pointerDec size) .
>            pointerInc (size * (length ws - 1))
>
> readSeqAt :: Int -> (Pointer a -> IO b) -> Int -> Pointer a -> IO [b]
> readSeqAt size readFun count
>       = maplRev readFun . take count .
>               iterate (pointerDec size) . pointerInc ((count-1)*size)
>
> readSeqAtWhile :: Int -> (Pointer a -> IO b) -> (b -> Bool) -> Pointer a -> IO [b]
> readSeqAtWhile size readFun test
>       = maplWhile test readFun . iterate (pointerInc size)


The Pointer data type is build upon the untyped Ptr data type.

@> newtype Ptr           = Ptr ForeignObj Ptr

> type FinalizerPtr     = Addr          -- Ptr -> IO ()
>
> makePtr :: FinalizerPtr -> Addr -> IO Ptr
> makePtr free addr
>       = do foreign <- makeForeignObj addr free
>            let p = Ptr foreign p
>            return p
>
> setFinalizerPtr free ptr
>       = return ()           -- unsupported yet
>
> setRootPtr root (Ptr p _)
>       = Ptr p root
>
> getRootPtr (Ptr p root)
>       = root

Memory allocation creates a @CPointer@.

Use @stackFrame@ when you are sure to deallocate some memory.
This kind of memory allocation will probably be optimized.

> allocStack    = allocFrame
>
> stackFrame :: Int -> (Pointer a -> IO b) -> IO b
> stackFrame size f
>       = do p <- allocFrame size
>            f p `always` freeMemory p


Special free routines for pointers. Use them to manually free pointers.

> freeMemory            = freeWithC finalFreeMemory
> freeInterfacePointer  = freeWithC finalFreeObject
> freeBSTR              = freeWithC finalFreeBSTR
>
>
> freeWithC :: Finalizer a -> Pointer a -> IO ()
> freeWithC final p@(Pointer (Ptr f root))
>       = do addr <- unsafeAddrFromPointer p
>            primFinalize final addr
>            writeForeignObj f addrNULL
>
> freeWith :: (Pointer a -> IO ()) -> Pointer a -> IO ()
> freeWith free p@(Pointer (Ptr f root))
>       = do free p
>            writeForeignObj f addrNULL
                

\end{document}
